// implementa l'algoritmo di Sammon (IEEE trans. con compuiters vol. C-18, p. 401

import java.util.*;
import java.io.*;
import java.awt.*;

class HiPointDataset extends Vector {
  public double[] mins,maxs,normalized_averages,normalized_variances,averages,variances;
  public int dimensions,best_one,second_one;
  double[] D2data; // matrice ausiliaria
  double[]distXi; // i d* dell'articolo
  int N;
  int nrPoints;
  double sumDistXi = 0; /* il c dell'articolo */
  void compute_dimension_stats() throws IllegalFormatException
    {
      int i,j;
      double t;
      dimensions=myElementAt(0).hicoords.length;
      mins=new double[dimensions];
      maxs=new double[dimensions];
      normalized_averages=new double[dimensions];
      normalized_variances=new double[dimensions];
      averages=new double[dimensions];
      variances=new double[dimensions];
      for(i=0; i!=dimensions; i++)
        {
          mins[i]=myElementAt(0).hicoords[i];
          maxs[i]=myElementAt(0).hicoords[i];
          averages[i]=myElementAt(0).hicoords[i];
          variances[i]=0;
          normalized_variances[i]=0;
        }
      for(i=1; i<size(); i++)
        {
          if(dimensions!=myElementAt(i).hicoords.length)
            throw(new IllegalFormatException());
          for(j=0; j!=dimensions; j++)
            {
              t=myElementAt(i).hicoords[j];
              if(t<mins[j]) mins[j]=t;
              if(t>maxs[j]) maxs[j]=t;
              averages[j]+=t;
            }
        }
      best_one=0;
      second_one=0;
      for(j=0; j!=dimensions; j++)
        {
          averages[j]=averages[j]/(double)size();
          normalized_averages[j]=(averages[j]-mins[j])/(maxs[j]-mins[j]);
        }
      for(j=0; j!=dimensions; j++)
        {
          for(i=0; i<size(); i++)
           {
             t=myElementAt(i).hicoords[j]-averages[j];
             variances[j]+=t*t;
             t=(myElementAt(i).hicoords[j]-mins[j])/(maxs[j]-mins[j])-normalized_averages[j];
             normalized_variances[j]+=t*t;
           }
         variances[j]=variances[j]/(double)size();
         normalized_variances[j]=normalized_variances[j]/(double)size();
         System.out.println("Dimension "+j+" average "+averages[j]+" variance="+variances[j]+" min="+mins[j]+" max="+maxs[j]);
         System.out.println("   Normalized average "+normalized_averages[j]+" variance="+normalized_variances[j]);
        }
      for(j=0; j!=dimensions; j++)
        {
         System.out.println(normalized_variances[j]+"["+j+"] vs "+normalized_variances[best_one]+"["+best_one+"]");
         if(normalized_variances[j]>=normalized_variances[best_one])
           {
             second_one=best_one; best_one=j;
             System.out.println("best_one="+best_one+" second_one="+second_one);
           }
         else if(normalized_variances[j]>=normalized_variances[second_one])
           {
             second_one=j;
             System.out.println("best_one="+best_one+" second_one="+second_one);
           }
       }
      System.out.println("best_one="+best_one+" second_one="+second_one);
      for(i=0; i!=size(); i++)
        {
          myElementAt(i).x=myElementAt(i).hicoords[best_one];
          myElementAt(i).y=myElementAt(i).hicoords[second_one];
        }
    }
  void read_dataset(InputStream fi) throws IOException,IllegalFormatException
    {
      Vector buff;
      buff=new Vector();
      StreamTokenizer w=new StreamTokenizer(fi);
      w.parseNumbers(); w.slashStarComments(true);
      w.eolIsSignificant(true);
      while(w.nextToken()!=StreamTokenizer.TT_EOF)
        {
          if(w.ttype==StreamTokenizer.TT_EOL)
            {
              addElement(new HiPoint(buff));
              buff=new Vector();
            }
          else if(w.ttype==StreamTokenizer.TT_NUMBER)
            {
              buff.addElement(new Double(w.nval));
            }
        }
      D2data=null;
      N=-1;
      nrPoints=size();
      compute_dimension_stats();
    }
  HiPoint myElementAt(int i)
    {
      return (HiPoint)(super.elementAt(i));
    }
  HiPointDataset(String s) throws IOException,IllegalFormatException
      {
        FileInputStream fi=new FileInputStream(s);
        read_dataset(fi);
        fi.close();
      }
  public double iterate_sammon(TextField iterate1,TextField iterate2,TextField r1,TextField r2)
    {
      /* la procedura di Sammon vera e propria... ritorna il nuovo errore */
      
	int		i,j,h,k,XdistIndex,YdistIndex,p,q;   	// some indexes
	double[]	x2 = new double[2];  		// x[i]  one point of the 2-dataset
	double[]	y2 = new double[2];  		// y[i]  one point of yhe 2-dataset
	double		error = 0;  		// Sammon error
	double		errorInitial;  		

	double		sumSquareMeanError = 0; 
	double		MF = 0.3;   // Magic Factor

	double		delta_pq = 0; 
	double 		sumInDer1 = 0;
	double 		sumInDer2 = 0;
	double 		C = 0;

	StringBuffer sb;

        if(D2data==null)
           {
	     D2data= new double[ 2*nrPoints ];       // target
	     N=0;
	     for (i=1; i<=nrPoints ;i++)  //nr of interpoint distances to calculate
	       N += i - 1;
	     System.out.println("Computing N="+N+" interpoint distances");    
	     r1.setText("Computing real interpoint distances");
             r2.setText("Preparing data...");	
	     distXi = new double[N];
	     sb=new StringBuffer();
	     r2.setText("");
	     XdistIndex = 0;
	     for (i=0;i<nrPoints-1;i++) 
	       {  // tiene il conto delle iterazioni  di2, di3, di4, ...
	         if((nrPoints<=80) || (i%(1+nrPoints/80)==0))
	           {
	             sb.append("|");
	             r2.setText(new String(sb));
	           }
	         for (j=i+1;j<nrPoints;j++)
	           {  // tiene il conto delle iterazioni  d1j, d1j, d1j, ...
		     distXi[XdistIndex] = dist(myElementAt(i).hicoords,myElementAt(j).hicoords);
		     if (distXi[XdistIndex]==0) 
		       distXi[XdistIndex]=1e-7; /* epsilon per evitare distanze nulle */
		     sumDistXi += distXi[XdistIndex];   // I need of this for calculating the error  
		     XdistIndex++;
                   }
	       }
           }
	r1.setText("Computing apparent interpoint distances");
        r2.setText("Preparing data...");	
	double[]distYi = new double[N];   //ricontrollare tutti gli indici dei vettori...!!!!
	for (i=0;i<D2data.length;i++) 
          D2data[i] = ((i%2)==0)?myElementAt(i/2).x:myElementAt(i/2).y;
	/* for (i=0;i<D2data.length;i++) 
	    System.out.println("D2data["+i+"]-> "+D2data[i]); */
	YdistIndex = 0;
	sb=new StringBuffer();
	r2.setText("");
	//for (i=0;i<NDData.length;i++) 
	    //System.out.println("NDData["+i+"]-> "+ NDData[i]);
	for (i=0;i<nrPoints-1;i++) 
	  {  // tiene il conto delle iterazioni  di2, di3, di4, ...
	    if((nrPoints<=80) || (i%(1+nrPoints/80)==0))
	      {
	        sb.append("|");
	        r2.setText(new String(sb));
	      }
            for (h=posIntwoDData(i), k=0; h<(posIntwoDData(i)+2); k++,h++) 		    
              x2[k] = D2data[h];
	    for (j=i+1;j<nrPoints;j++)
	      {
		for (h=posIntwoDData(j), k=0; h<(posIntwoDData(j)+2);k++,h++) 		    
		  y2[k] = D2data[h];  
		distYi[YdistIndex] = dist(x2,y2);
		if (distYi[YdistIndex]==0) distYi[YdistIndex]=1e-7; /* epsilon per evitare distanze nulle */
		YdistIndex++;    
	      }	    	
	  }
	r1.setText("Computing initial error...");
	r2.setText("");
	sb=new StringBuffer();
        System.out.println("Computing initial error...");
	for (i=0;i<distXi.length;i++) 
	  {
	    if(i%(distXi.length/80)==0)
	      {
		sb.append("|");
	        r2.setText(new String(sb));
	      }
/*	    System.out.println("distXi["+i+"]-> "+ distXi[i]);
	    System.out.println("distYi["+i+"]-> "+ distYi[i]); */
	    sumSquareMeanError += ((Math.pow((distXi[i]-distYi[i]),2))/distXi[i]);
	  }
	errorInitial = error = (1/sumDistXi)*sumSquareMeanError;
	System.out.println("sumDistXi-> "+sumDistXi);
	System.out.println("error-> "+error);
	C = (-2/sumDistXi);
	System.out.println("C-> "+C);
	r1.setText("Transforming...");
	r2.setText("");
	sb=new StringBuffer();
	iterate2.setText("Error: "+error);
	System.out.println("Transforming...");
	sumInDer1 = 0;
	sumInDer2 = 0;
	delta_pq = 0;
        for (p=0;p<nrPoints;p++)  
          {  //for 0
	    if((nrPoints<=80) || (p%(1+nrPoints/80)==0))
	      {
	        sb.append("|");
	        r2.setText(new String(sb));
		System.out.print(".");
	      }
	    for (q=0;q<2;q++)  
	      {  //for 1
	        sumInDer1 = 0; /* per me vanno all'interno del ciclo... RP */
	        sumInDer2 = 0; /* RP */
	        for (j=0;j<nrPoints;j++) 
	          {
		    if (j!=p) 
		      {
			// System.out.println("dx_ij(p,j,nrPoints)-> "+dx_ij(p,j,nrPoints)+"  p-> "+p+"  j-> "+j);
			// System.out.println("distXi="+distXi[dx_ij(p,j,nrPoints)]+", distYi="+distYi[dx_ij(p,j,nrPoints)]);
			//System.out.println("matrix2arrayPos(p,q,dimensions)-> "+matrix2arrayPos(p,q,dimensions)+"  q-> "+q);
			sumInDer1 += ((distXi[dx_ij(p,j,nrPoints)]-distYi[dx_ij(p,j,nrPoints)]) / 
			              (distXi[dx_ij(p,j,nrPoints)] * distYi[dx_ij(p,j,nrPoints)]) ) * 
			              (D2data[matrix2arrayPos(p,q,2)] - D2data[matrix2arrayPos(j,q,2)]  );
			sumInDer2 += ( 1 / (distXi[dx_ij(p,j,nrPoints)] * distYi[dx_ij(p,j,nrPoints)]))* 
			             ( (distXi[dx_ij(p,j,nrPoints)]-distYi[dx_ij(p,j,nrPoints)]) - 
			                (((Math.pow((D2data[matrix2arrayPos(p,q,2)] - D2data[matrix2arrayPos(j,q,2)]),2) 
			                 / distYi[dx_ij(p,j,nrPoints)])) 
			                * (1+((distXi[dx_ij(p,j,nrPoints)]-distYi[dx_ij(p,j,nrPoints)])/distYi[dx_ij(p,j,nrPoints)]))));
		            // System.out.println("sumInDer1="+sumInDer1+"; sumInDer2="+sumInDer2);
		      }
		  }
	        delta_pq = (  (C*sumInDer1) / (Math.abs(C*sumInDer2))  );
	        // System.out.println("sumInDer1="+sumInDer1+"; sumInDer2="+sumInDer2);
		// System.out.println("D2data["+matrix2arrayPos(p,q,2)+"]="+D2data[matrix2arrayPos(p,q,2)]);
		D2data[matrix2arrayPos(p,q,2)] -= (MF * delta_pq);  // Ypq(m+1)
		// System.out.println("; delta_pq->  "+ delta_pq+"; result="+D2data[matrix2arrayPos(p,q,2)]);
	      }  // for 1
	  }  //for 0
	System.out.println("");
	r2.setText("");
	r1.setText("Recomputing error");
//begin error calculation
	YdistIndex = 0;
	for (i=0;i<nrPoints-1;i++) 
	  {  // tiene il conto delle iterazioni  di2, di3, di4, ...
	    for (h=posIntwoDData(i), k=0; h<(posIntwoDData(i)+2); k++,h++) 		    
	      x2[k] = D2data[h];
	    for (j=i+1;j<nrPoints;j++)
	      {
	        for (h=posIntwoDData(j), k=0; h<(posIntwoDData(j)+2);k++,h++) 		    
	          y2[k] = D2data[h];  
	        distYi[YdistIndex] = dist(x2,y2);
		YdistIndex++;    
	      }
	  }    
	sumSquareMeanError = 0;
	for (i=0;i<distXi.length;i++) {
	    sumSquareMeanError += ((Math.pow((distXi[i]-distYi[i]),2))/distXi[i]);
	}
	error = (1/sumDistXi)*sumSquareMeanError;
	r1.setText("Done.");
	System.out.println("RESULTING ERROR-> "+ error);
	for (i=0;i<D2data.length;i+=2) 
	  {
	    myElementAt(i/2).x=D2data[i];
	    myElementAt(i/2).y=D2data[i+1];
	    // System.out.print(D2data[i]+"  ");
	    // System.out.println(D2data[i+1]);		        	        
          }
	if (error>errorInitial) 
	  {
            System.out.println("WARNING: error worsenend");
	    r2.setText("WARNING: error worsened");
          }
//end error calculation 
	/* for (i=0;i<D2data.length;i+=2) {
	    System.out.print(D2dataFinal[i]+"  ");
	    System.out.println(D2dataFinal[i+1]);		    	
	} */
       return error;
    }
    
    int posIntwoDData(int elementNr) {
	return ((elementNr)*2);    
    }

    int matrix2arrayPos(int row, int col, int nrColumns)  { // da (0,0) (row,col) -> elementNr
	return ((row)*nrColumns + col);
    
    }

     double dist(double[] x, double[] y) {  //distance between two vectors of the space
	int i=0;			   // same length is assumed!!!!
	double sum=0;
	
	    
	for (i=0;i<x.length;i++) 
	    sum += Math.pow((x[i]-y[i]),2);

	return Math.sqrt(sum);
    
    }
    
    int dx_ij(int i, int j, int nrPoints)  {   // d_ij (i,j, nrPoints) -> elementNr
						// iff i<j tuttoOK
	int sumPos = 0;
	
	if (i<j) {	
	    for (int k=1; k<i+1; k++) 
		sumPos += nrPoints - k;
	    sumPos += j-i-1;
	    return sumPos;
	    
	} else if (i>j) {
	    sumPos = dx_ij(j,i,nrPoints); // controllare...!! comunque il caso i=j e' escluso dal chiamante
	    return sumPos;
	    
	}
	return -1;
    }    
    

}

